# -*- coding: utf_8 -*-
"""Logger Config."""
import sys
import logging

from colorama import Back, Fore, Style


class ColorLogsWrapper(object):
    COLOR_MAP = {
        'debug': Fore.CYAN,
        'info': Fore.BLUE,
        'warning': Fore.YELLOW,
        'error': Fore.RED,
        'critical': Back.RED,
    }

    def __init__(self, logger):
        self.logger = logger

    def __getattr__(self, attr_name):
        """Getattr."""
        if attr_name == 'info':
            attr_name = 'info'
        if attr_name not in 'debug info warning error critical':
            return getattr(self.logger, attr_name)
        log_level = getattr(logging, attr_name.upper())
        # mimicking logging/__init__.py behaviour
        if not self.logger.isEnabledFor(log_level):
            return

        def wrapped_attr(msg, *args, **kwargs):
            style_prefix = self.COLOR_MAP[attr_name]
            msg = style_prefix + str(msg) + Style.RESET_ALL
            # We call _.log directly to not increase the callstack
            # so that Logger.findCaller extract the corrects filename/lineno
            return self.logger._log(log_level, msg, args, **kwargs)
        return wrapped_attr


def init_logger(module_name) -> logging.Logger:
    """Setup logger."""
    log_format = '%(message)s'
    logging.basicConfig(stream=sys.stderr,
                        format=log_format,
                        level=logging.INFO)
    logger_obj = logging.getLogger(module_name)
    logger = ColorLogsWrapper(logger_obj)
    return logger



# -*- coding: utf_8 -*-
"""JSON output format."""
import json


def json_output(outfile, scan_results, version):
    """JSON Output."""
    scan_results['njsscan_version'] = version
    jout = json.dumps(
        scan_results,
        sort_keys=True,
        indent=2,
        separators=(',', ': '))
    if outfile:
        with open(outfile, 'w') as of:
            of.write(jout)
    else:
        print(jout)
    return jout



# -*- coding: utf_8 -*-
"""The nodejsscan cli: njsscan."""
from libsast import Scanner

from njsscan import settings
from njsscan.utils import (
    get_config,
    read_missing_controls,
)


class NJSScan:
    def __init__(self, paths, json, check_controls) -> None:
        conf = get_config(paths)
        self.check_controls = check_controls
        self.options = {
            'match_rules': settings.PATTERN_RULES_DIR,
            'sgrep_rules': settings.SGREP_RULES_DIR,
            'sgrep_extensions': conf['nodejs_extensions'],
            'match_extensions': conf['template_extensions'],
            'ignore_filenames': conf['ignore_filenames'],
            'ignore_extensions': conf['ignore_extensions'],
            'ignore_paths': conf['ignore_paths'],
            'ignore_rules': conf['ignore_rules'],
            'show_progress': not json,
        }
        self.paths = paths
        self.result = {
            'templates': {},
            'nodejs': {},
            'errors': [],
        }

    def scan(self) -> dict:
        """Start Scan."""
        scanner = Scanner(self.options, self.paths)
        result = scanner.scan()
        if result:
            self.format_output(result)
        return self.result

    def format_output(self, results) -> dict:
        """Format to njsscan friendly output."""
        self.format_sgrep(results['semantic_grep'])
        self.format_matches(results['pattern_matcher'])
        self.post_ignore_rules()
        self.post_ignore_files()

    def missing_controls(self, result):
        """Check for missing controls."""
        controls = read_missing_controls()['controls']
        result_keys = result['nodejs'].keys()
        for rule_id in settings.GOOD_CONTROLS_ID:
            if rule_id in result_keys:
                # Good! Control is present
                del result['nodejs'][rule_id]
            else:
                if self.check_controls:
                    result['nodejs'][rule_id] = controls[rule_id]
                else:
                    continue

    def format_sgrep(self, sgrep_output):
        """Remove metavars from sgrep output."""
        self.result['errors'] = sgrep_output['errors']
        for rule_id in sgrep_output['matches']:
            for finding in sgrep_output['matches'][rule_id]['files']:
                finding.pop('metavars', None)
        self.result['nodejs'] = sgrep_output['matches']
        self.missing_controls(self.result)

    def format_matches(self, matcher_out):
        """Format Pattern Matcher output."""
        self.result['templates'] = matcher_out

    def post_ignore_rules(self):
        """Ignore findings by rules."""
        for rule_id in self.options['ignore_rules']:
            if rule_id in self.result['nodejs']:
                del self.result['nodejs'][rule_id]
            if rule_id in self.result['templates']:
                del self.result['templates'][rule_id]

    def post_ignore_files(self):
        """Ignore file by rule."""
        del_keys = set()
        for rule_id, details in self.result['nodejs'].items():
            files = details.get('files')
            if not files:
                continue
            tmp_files = files.copy()
            for file in files:
                mstr = file.get('match_string')
                if 'ignore:' in mstr and rule_id in mstr:
                    tmp_files.remove(file)
                if len(tmp_files) == 0:
                    del_keys.add(rule_id)
            details['files'] = tmp_files
        # Remove Rule IDs marked for deletion.
        for rid in del_keys:
            if rid in self.result['nodejs']:
                del self.result['nodejs'][rid]



#!/usr/bin/env python
# -*- coding: utf_8 -*-

__title__ = 'njsscan'
__authors__ = 'Ajin Abraham'
__copyright__ = 'Copyright 2020 Ajin Abraham, OpenSecurity'
__version__ = '0.1.9'
__version_info__ = tuple(int(i) for i in __version__.split('.'))
__all__ = [
    '__title__',
    '__authors__',
    '__copyright__',
    '__version__',
    '__version_info__',
]



# -*- coding: utf_8 -*-
"""Sarif output format.

Based on https://github.com/microsoft/bandit-sarif-formatter/
blob/master/bandit_sarif_formatter/formatter.py

Copyright (c) Microsoft.  All Rights Reserved.
MIT License

Copyright (c) Microsoft Corporation.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE
"""
from datetime import datetime
from pathlib import PurePath
import urllib.parse as urlparse

import sarif_om as om

from jschema_to_python.to_json import to_json


TS_FORMAT = '%Y-%m-%dT%H:%M:%SZ'


def level_from_severity(severity):
    if severity == 'ERROR':
        return 'error'
    elif severity == 'WARNING':
        return 'warning'
    elif severity == 'INFO':
        return 'note'
    else:
        return 'none'


def to_uri(file_path):
    pure_path = PurePath(file_path)
    if pure_path.is_absolute():
        return pure_path.as_uri()
    else:
        posix_path = pure_path.as_posix()  # Replace backslashes with slashes.
        return urlparse.quote(posix_path)  # %-encode special characters.


def get_rule_name(rule_id):
    normalized = []
    noms = rule_id.split('_')
    for nom in noms:
        normalized.append(nom.capitalize())
    return ''.join(normalized)


def add_results(scan_results, run):
    if run.results is None:
        run.results = []
    res = {}
    res.update(scan_results.get('nodejs', []))
    res.update(scan_results.get('templates', []))
    rules = {}
    rule_indices = {}

    for rule_id, issue_dict in res.items():
        if 'files' not in issue_dict:
            # no missing controls in sarif
            continue
        result = create_result(rule_id, issue_dict, rules, rule_indices)
        run.results.append(result)

    if len(rules) > 0:
        run.tool.driver.rules = list(rules.values())


def create_result(rule_id, issue_dict, rules, rule_indices):
    if rule_id in rules:
        rule = rules[rule_id]
        rule_index = rule_indices[rule_id]
    else:
        doc = 'https://ajinabraham.github.io/nodejsscan/#{}'.format(rule_id)
        rule = om.ReportingDescriptor(
            id=rule_id,
            name=get_rule_name(rule_id),
            help_uri=doc,
        )
        rule_index = len(rules)
        rules[rule_id] = rule
        rule_indices[rule_id] = rule_index

    locations = []
    for item in issue_dict['files']:
        physical_location = om.PhysicalLocation(
            artifact_location=om.ArtifactLocation(
                uri=to_uri(item['file_path'])),
        )
        physical_location.region = om.Region(
            start_line=item['match_lines'][0],
            end_line=item['match_lines'][1],
            start_column=item['match_position'][0],
            end_column=item['match_position'][1],
            snippet=om.ArtifactContent(text=item['match_string']),
        )
        locations.append(om.Location(physical_location=physical_location))

    return om.Result(
        rule_id=rule.id,
        rule_index=rule_index,
        message=om.Message(text=issue_dict['metadata']['description']),
        level=level_from_severity(issue_dict['metadata']['severity']),
        locations=locations,
        properties={
            'owasp': issue_dict['metadata']['owasp'],
            'cwe': issue_dict['metadata']['cwe'],
        },
    )


def sarif_output(outfile, scan_results, njsscan_version):
    log = om.SarifLog(
        schema_uri=('https://raw.githubusercontent.com/oasis-tcs/'
                    'sarif-spec/master/Schemata/sarif-schema-2.1.0.json'),
        version='2.1.0',
        runs=[
            om.Run(
                tool=om.Tool(driver=om.ToolComponent(
                    name='nodejsscan',
                    information_uri='https://github.com/ajinabraham/njsscan',
                    semantic_version=njsscan_version,
                    version=njsscan_version),
                ),
                invocations=[
                    om.Invocation(
                        end_time_utc=datetime.utcnow().strftime(TS_FORMAT),
                        execution_successful=True,
                    ),
                ],
            ),
        ],
    )
    run = log.runs[0]
    add_results(scan_results, run)
    json_out = to_json(log)
    if outfile:
        with open(outfile, 'w') as of:
            of.write(json_out)
    else:
        print(json_out)
    return json_out



# -*- coding: utf_8 -*-
"""Sonarqube output format."""

from njsscan.formatters.json import json_output


def get_sonarqube_issue(njsscan_issue):
    sonarqube_severity_mapping = {
        'ERROR': 'CRITICAL',
        'WARNING': 'MAJOR',
        'INFO': 'MINOR',
    }
    secondary_locations = []
    issue_data = njsscan_issue['metadata']
    # Handle missing controls
    if not njsscan_issue.get('files'):
        primary_location = None
    else:
        for ix, file in enumerate(njsscan_issue['files']):
            text_range = {
                'startLine': file['match_lines'][0],
                'endLine': file['match_lines'][1],
                'startColumn': file['match_position'][0],
                'endColumn': file['match_position'][1],
            }
            location = {
                'message': issue_data['description'],
                'filePath': file['file_path'],
                'textRange': text_range,
            }
            if ix == 0:
                primary_location = location
            else:
                secondary_locations.append(location)
    issue = {
        'engineId': 'njsscan',
        'type': 'VULNERABILITY',
        'severity': sonarqube_severity_mapping[issue_data['severity']],
        'primaryLocation': primary_location,
    }
    if secondary_locations:
        issue['secondaryLocations'] = secondary_locations
    return issue


def sonarqube_output(outfile, scan_results, version):
    """Sonarqube JSON Output."""
    sonarqube_issues = []
    for i in ['nodejs', 'templates']:
        for k, v in scan_results[i].items():
            issue = get_sonarqube_issue(v)
            issue['ruleId'] = k
            sonarqube_issues.append(issue)
    sonarqube_report = {
        'issues': sonarqube_issues,
    }
    return json_output(outfile, sonarqube_report, version)



#!/usr/bin/env python
# -*- coding: utf_8 -*-
"""The nodejsscan cli: njsscan."""
import argparse
import sys

from njsscan import __version__
from njsscan.njsscan import NJSScan
from njsscan.formatters import (
    cli,
    json,
    sarif,
    sonarqube,
)


def handle_exit(results, exit_warn):
    """Handle Exit."""
    combined = {}
    if results.get('nodejs'):
        combined.update(results['nodejs'])
    if results.get('templates'):
        combined.update(results['templates'])
    for meta in combined.values():
        severity = meta['metadata']['severity']
        ewarn = severity == 'WARNING' and exit_warn
        if severity == 'ERROR' or ewarn:
            sys.exit(1)
    sys.exit(0)


def main():
    """Main CLI."""
    parser = argparse.ArgumentParser()
    parser.add_argument('path',
                        nargs='*',
                        help=('Path can be file(s) or '
                              'directories with source code'))
    parser.add_argument('--json',
                        help='set output format as JSON',
                        action='store_true')
    parser.add_argument('--sarif',
                        help='set output format as SARIF 2.1.0',
                        action='store_true')
    parser.add_argument('--sonarqube',
                        help='set output format compatible with SonarQube',
                        action='store_true')
    parser.add_argument('-o', '--output',
                        help='output filename to save the result',
                        required=False)
    parser.add_argument('--missing-controls',
                        help='enable missing security controls check',
                        action='store_true',
                        required=False)
    parser.add_argument('-w', '--exit-warning',
                        help='non zero exit code on warning',
                        action='store_true',
                        required=False)
    parser.add_argument('-v', '--version',
                        help='show njsscan version',
                        required=False,
                        action='store_true')
    args = parser.parse_args()
    if args.path:
        is_json = args.json or args.sonarqube or args.sarif
        scan_results = NJSScan(
            args.path,
            is_json,
            args.missing_controls,
        ).scan()
        if args.sonarqube:
            sonarqube.sonarqube_output(args.output, scan_results, __version__)
        elif args.json:
            json.json_output(args.output, scan_results, __version__)
        elif args.sarif:
            sarif.sarif_output(args.output, scan_results, __version__)
        else:
            cli.cli_output(args.output, scan_results, __version__)
        handle_exit(scan_results, args.exit_warning)

    elif args.version:
        cli.print_tool_info(__version__)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



# -*- coding: utf_8 -*-
"""CLI njsscan output format."""
from tabulate import tabulate

from njsscan.logger import init_logger

logger = init_logger(__name__)


def print_tool_info(ver):
    """Tool info."""
    tool_str = '\nnjsscan: v{} | Ajin Abraham | opensecurity.in'.format(ver)
    logger.info(tool_str)
    return tool_str


def format_table(rule_id, details):
    """Get CLI friendly format."""
    tbl_fmt = 'fancy_grid'
    items = []
    items.append(['RULE ID', rule_id])
    for meta, value in details['metadata'].items():
        if meta == 'id':
            continue
        meta_format = meta.upper().replace('_', '')
        items.append([meta_format, value])
    # Get files
    fstore = []
    files = details.get('files')
    if files:
        for match in files:
            file_path = match['file_path']
            fstore.append(['File', file_path])
            position = match['match_position']
            pos = f'{position[0]} - {position[1]}'
            fstore.append(['Match Position', pos])
            lines = match.get('match_lines')
            line = (lines[0] if lines[0] == lines[1]
                    else f'{lines[0]}: {lines[1]}')
            fstore.append(['Line Number(s)', line])
            match_string = match['match_string']
            if isinstance(match_string, list):
                match_string = '\n'.join(ln.strip() for ln in match_string)
            fstore.append(['Match String', match_string])
        if fstore:
            files_tbl = tabulate(fstore, tablefmt=tbl_fmt)
            items.append(['FILES', files_tbl])
    return tabulate(items, tablefmt=tbl_fmt)


def cli_output(outfile, scan_results, version):
    """Format output printing."""
    tool = print_tool_info(version)
    if not (scan_results['nodejs'] or scan_results['templates']):
        logger.info('No issues found.')
        return []
    scan_results.pop('errors', None)
    buffer = []
    for out in scan_results:
        for rule_id, details in scan_results[out].items():
            formatted = format_table(rule_id, details)
            buffer.append(formatted)
            severity = details['metadata']['severity'].lower()
            if not outfile:
                if severity == 'error':
                    logger.error(formatted)
                elif severity == 'warning':
                    logger.warning(formatted)
                else:
                    logger.info(formatted)
    if outfile and buffer:
        buffer.insert(0, tool)
        outdata = '\n'.join(buffer)
        with open(outfile, 'w') as of:
            of.write(outdata)
    return buffer
